//  RW Foundations
// ================
// ______
// -The RW Foundational Libraries

// Filename: ./include/rw/rw.hpp
// -Declarations of the baseline RW library

#pragma once

#include "incls_all.hpp"

namespace rw {

//------------------------------------------------------------------------------
//    RW
//------------------------------------------------------------------------------
/** The master base class from which all other tagged rw:: classes derive
 * @note:
 */
class RW {
 public:
  /** Construct a @c RW object
   * @note default ctor
   */
  RW();

  /**
   * @param handle_tag
   * @param type
   */
  RW(std::string const& handle_tag, Type const type = Type::Foo);

  ~RW();

  /** RW object's Hip @a handle (it's 'name tag')
   * @return the object's handle
   */
  auto hdl() const -> std::string;

  /** RW object's Hip @a ID
   * @return the object's id
   */
  auto id() const -> unsigned int;

  /** RW object's Hip @a address
   * @return the object's address (it's pointer)
   */
  auto addr() const -> RW*;

  /** RW object's Hip @a type
   * @return
   */
  auto typ() const -> Type;

  /** Returns the @c RW object's @c Hip @a 'tag' (the full data structure)
   * @note: sending/receiving these is our primary means for tracing callgraphs,
   * and for access-control
   * @return read-only copy of the @c RW object's @c Hip member field
   */
  auto t() const -> Hip const;

  /**
   * @return
   */
  inline auto get_stamps() const -> Stampstack const { return stamps_; }

 protected:
  /** Adds a @c Callstamp for @a func_nm function into the @c Stampstack member
   * field
   * @param tag
   * @param func_nm
   * @note: this capability is switched by the @c do_stamps() member function
   */
  void stamp(Hip const& tag, std::string const& func_nm);
  // self-use wrapper
  void stamp(std::string const& func_nm);

  /** Controls whether to use call-stamping or not
   * @param tag
   * @param flag
   */
  void do_stamps(Hip const& tag, bool const flag);

 private:
  /** ID's & ptrs are unique, but handles @a may be identical among same types.
   * @note: eg; {sumomo.0}, {sumomo.1}, {arigato.2}, {dollhouse.3}, {arigatoB.4}
   */
  Hip const hip_;

  Stampstack stamps_;
  bool       do_stamps_;

  /** Incremented by one for each new instance of this base class (when created)
   */
  static unsigned long rw_id;
};

//------------------------------------------------------------------------------
//    RW_slave
//------------------------------------------------------------------------------
/** Provides @a suited-to-the-class features for all the *slave object* classes
 * @note intermediate base class from which all the *slave object* classes
 * derive
 */
class RW_slave : public RW {
 public:
  RW_slave() = delete;  // no default ctor

  /**
   * @param masters
   * @param handle
   */
  RW_slave(std::vector<Type> const& masters, std::string const& handle);

  /**
   * @param masters
   * @param handle
   * @param type
   */
  RW_slave(std::vector<Type> const& masters,
           std::string const&       handle,
           Type const               type);

  /** Attempt to bind this *slave object* to a suitable @a Master
   * @param[in] tag @c Hip of the would-be @a Master
   * @return true if the bind was successful
   */
  auto bind_master(Hip const& tag) -> bool;

  /** This *slave object* belongs to this @a Master
   * @return const copy of this *slave object's* @c Hip member field 'master_'
   * @note: imprinted & fixed within @c bind_master()
   * @note: the acceptable types of @a Master 's for this *slave object* class
   * are specified ahead of time (in the derived *slave object's* member init
   * list), and fixed within @c master_types()  (called during this *slave
   * object* class' constructor)
   */
  inline auto master() -> Hip const {
    stamp("master()");

    return master_;
  }

 protected:
  /** Stores the collection which determines what kind(s) of @a Master (s) this
   * *slave object* accepts binding to (one-only from the list at any time,
   * however)
   * @param[in] types a container of @c rw::Type 's acceptable as @a Master
   * @note called during this *slave object* class' constructor
   */
  void master_types(std::vector<Type> const& masters);

 private:
  vector<Type> master_types_;
  Hip          master_;
};

}  // namespace rw

//    --=^=--

// "So is it true Sempai? Are robowaifus really better than any human woman?"

// rw_foundations_lib.v220108
// Copyright (2022)
// License MIT(Expat) https://opensource.org/licenses/MIT
