//   RW Dollnet
// ==============
// External, Inter-Networking Library
// -Part of the RW Foundations suite of libraries

// Filename: ./src/dollhouse/net/dollnet_util.cpp
// -Definitions of some baseline RW Dollnet utility functions

///////
// Dolldrop/Dollnet is meant for network-enabled RW projects only
#if __has_include(<curl_easy.h>)

#include "incls_all.hpp"

#include "curl_easy.h"

namespace rw::dollhouse::net {

//------------------------------------------------------------------------------

constexpr int tot_ticks{42};  // how wide the progress meter bar is (in chars)
Time          start;          // will start the instant before an xfer attempt

constexpr float KiB{1'024.f};
constexpr float MiB{KiB * KiB};

//------------------------------------------------------------------------------
/**
 */
enum class Dir {
  up,
  dn,
};

//------------------------------------------------------------------------------
/** Initialize a @a tagfile file
 * @param tagfile_nm
 * @note: initial tag number defaults to @a '000', so @a '001' will be the first
 * (that is, the @a next) tag number delivered by @c filetag() whenever this
 * function call is required
 */
void mk_tagfile(std::string const& file_nm) {
  ofstream o_file{};
  o_file.open(file_nm);
  o_file << "000\n";
  o_file.close();
}

//------------------------------------------------------------------------------
/** File-based scheme for doling out incremental numbers and (optionally) a
 * specified file-extension, provided together as a single tag
 * @return the (next) file tag
 * @note: intended to serve as a filename suffix for differentiating subsequent
 * @c Dolldrop files
 */
std::string filetag() {
  string const file_nm{"./dolldrops/tagfile"};

  // attempt to open tagfile
  ifstream i_file{};
  i_file.open(file_nm);
  //
  if (! i_file) {  // create if needed
    mk_tagfile(file_nm);
    i_file.open(file_nm);
  }

  // capture tagfile's contents into vars
  int    cnt{0};
  string ext{""};
  i_file >> cnt >> ext;  // DESIGN: any validation req'd here ?
  i_file.close();

  // store incremented count back out to file for next dolldrop cycle
  ofstream o_file{};
  o_file.open(file_nm);
  o_file << setw(3) << setfill('0') << ++cnt << '\n'
         << ((ext != "") ? ext + '\n' : "");  // also write ext if present
  o_file.close();

  // create file tag from now-incremented tagfile data
  ostringstream tagss{};
  tagss << '_' << setw(3) << setfill('0') << cnt << ext;

  return tagss.str();
}

//------------------------------------------------------------------------------
/**
 * @param xfer_frac
 * @return
 */
std::string pct(float const xfer_frac) {
  auto          pct{xfer_frac * 100};
  ostringstream pctss;
  pctss << setw(6) << setfill(' ') << fixed << setprecision(1) << pct;

  return pctss.str();
}

//------------------------------------------------------------------------------
/**
 * @param xfer_frac
 * @param do_rev optional flag, reverse bar direction
 * @return
 */
std::string ticks(float const xfer_frac, bool const do_rev = false) {
  // part already done (the tickmarks)
  auto done = static_cast<int>(round(tot_ticks * xfer_frac));

  // part still remaining (the spaces)
  auto rem{tot_ticks - done};

  ostringstream tickss;
  if (do_rev)
    tickss << setw(rem) << setfill(' ') << "" << setw(done) << setfill('=')
           << "";
  else
    tickss << setw(done) << setfill('=') << "" << setw(rem) << setfill(' ')
           << "";

  return tickss.str();
}

//------------------------------------------------------------------------------
/**
 * @param xfer_tot
 * @param xfer_done
 * @param units
 * @return
 */
std::string m_of_n(curl_off_t const xfer_tot,
                   curl_off_t const xfer_done,
                   float const      units) {
  ostringstream m_of_nss;
  m_of_nss.imbue(locale{""});  // provides for digit separators
  m_of_nss << setw(5) << fixed << setprecision(1) << (xfer_done / units)
           << " of " << setw(5) << fixed << setprecision(1)
           << (xfer_tot / units);

  return m_of_nss.str();

  // https://en.cppreference.com/w/cpp/locale/locale
  // https://en.cppreference.com/w/cpp/io/basic_ios/imbue
}

//------------------------------------------------------------------------------
/**
 * @param durn
 * @return
 */
std::string hms_fmt(Durn durn) {  //

  auto const hrs = duration_cast<hours>(durn);
  durn -= hrs;
  auto const mins = duration_cast<minutes>(durn);
  durn -= mins;
  Secs_f const secs = durn;

  ostringstream hrss;
  ostringstream minss;

  auto const have_hrs{hrs.count() > 0};
  if (have_hrs)
    hrss << hrs.count() << 'h';

  auto const have_mins{mins.count() > 0};
  if (have_mins)
    minss << mins.count() << 'm';

  ostringstream hmss{};
  hmss << (have_hrs ? hrss.str() : "")    //
       << (have_mins ? minss.str() : "")  //
       << setw(4) << setfill(' ') << fixed << setprecision(1) << secs.count()
       << 's';

  return hmss.str();
}

//------------------------------------------------------------------------------
/**
 * @return
 * @note: simple wrapper providing now() time call
 */
std::string hms() {
  return hms_fmt(Call_clk::now() - start);
}

//------------------------------------------------------------------------------
/** The actual 'print the bar' code itself
 * @param xfer_tot
 * @param xfer_cmplt
 * @param dir
 * @param do_rev optional flag, reverse bar direction
 * @note: works for either upload or download
 */
void prt_bar(curl_off_t const xfer_tot,
             curl_off_t const xfer_cmplt,
             Dir const        dir,
             bool const       do_rev = false) {
  // cast prevents integer division
  auto xfer_frac = xfer_cmplt / static_cast<float>(xfer_tot);

  auto xfer_pct    = pct(xfer_frac);                     // percent string
  auto xfer_ticks  = ticks(xfer_frac, do_rev);           // tickmarks string
  auto xfer_m_of_n = m_of_n(xfer_tot, xfer_cmplt, KiB);  // bytes string

  cout << ((dir == Dir::up) ? "  u/l" : "  d/l") << xfer_pct << "% ["
       << xfer_ticks << "]  KiB : " << xfer_m_of_n << "  " << hms();
}

//------------------------------------------------------------------------------
/** Prints running count of bytes transferred during an unbounded (streaming)
 * connection, out to console
 * @param xfer_done
 * @param dir
 * @note: bytes output as measured in MiB
 */
void prt_bytes(curl_off_t const xfer_done, Dir const dir) {
  ostringstream raw_bytess{};
  raw_bytess.imbue(locale{""});  // provides for digit separators
  raw_bytess << ((dir == Dir::up) ? "  u/l" : "  d/l") << "   MiB : " << setw(7)
             << setfill(' ') << fixed << setprecision(2) << (xfer_done / MiB)
             << "   " << hms();

  cout << raw_bytess.str();
}

//------------------------------------------------------------------------------
/** Prints 'progress meter bars' for @b curl 's transfers, out to console
 *
 * This function acts as a callback, called at @b curl 's discretion while
 * transfers are in progress
 *
 * @param unused
 * @param dl_tot
 * @param dl_rcvd
 * @param ul_tot
 * @param ul_sent
 * @return default '0' value, unused
 * @note: transfer's status data are supplied by the @b curl library itself,
 * we are simply providing for the required form for it here and utilizing
 * them
 */
int prt_prog(Address,
             curl_off_t dl_tot  = 0,
             curl_off_t dl_rcvd = 0,
             curl_off_t ul_tot  = 0,
             curl_off_t ul_sent = 0) {  //

  bool have_ul{false}, have_dl{false};

  // UPLOAD
  if (ul_tot > 0) {  // have an upload to display?

    prt_bar(ul_tot, ul_sent, Dir::up);
    have_ul = true;
  } else if (ul_sent > 0) {  // unbounded ul going (streaming)?

    prt_bytes(ul_sent, Dir::up);
    have_ul = true;
  }
  if (have_ul)
    // line ending depends if we have a simultaneous dl (stacked bars, or not)
    cout << ((dl_tot > 0 || dl_rcvd > 0) ? "\n" : "\r");

  // DOWNLOAD
  if (dl_tot > 0) {  // have a download to display?

    prt_bar(dl_tot, dl_rcvd, Dir::dn, true);
    have_dl = true;
  } else if (dl_rcvd > 0) {  // unbounded dl going (streaming)?

    prt_bytes(dl_rcvd, Dir::dn);
    have_dl = true;
  }
  if (have_dl)
    // line ending depends if we have a simultaneous ul (upline + r, or just r)
    cout << ((ul_tot > 0 || ul_sent > 0) ? "\u008D\r" : "\r");

  if (have_ul || have_dl)
    // printing w/ flushing works smoother in this particular type case ('\r')
    cout << flush;

  // finally, transfer will abort w/o a '0' func ret
  return 0;
}

}  // namespace rw::dollhouse::net

#endif  // __has_include(<curl_easy.h>)

//    --=^=--

// "This apartment has super-high bandwidth illegal even to large corporations?"

// rw_dollhouse_lib.v220108
// Copyright (2022)
// License MIT (Expat) https://opensource.org/licenses/MIT
